import sys
import os
import random
import pygame
from pygame.locals import *


# Closure to hide the game's logic in an anonymous namespace
def GAME():
    # Initialize everything
    pygame.init()
    
    # Parameters, settings etc. for the entire game
    PARAM = {
        "ww"           : 640,                         # window width
        "wh"           : 480,                         # window height
        "bg_color"     : pygame.Color(0, 0, 0),       # window background color
        "speed_factor" : 1000.0,                      # correction for movement speed
        "init_balls"   : 2,                           # initial number of balls
        "ball_min"     : 10,                          # ball minimum radius
        "ball_max"     : 30,                          # ball maximum radius
        "bomb_size"    : 2,                           # bombs' radius
        "bomb_color"   : pygame.Color(255, 255, 255), # bombs' color
        "hero_size"    : 20,                          # hero's size
        "hero_color"   : pygame.Color(100, 255, 0),   # hero's color
        "hit_points"   : 11,                          # points for each hit
        "bomb_points"  : -1,                          # points for each bomb shot
        "score_font"   : pygame.font.Font(None, 30),  # font for the score
        "score_color"  : pygame.Color(100, 255, 0)    # color for the score
    }
    
    # Sprite containers
    SPRITES = {
        "balls" : pygame.sprite.Group(),        # group for all ball sprites
        "bombs" : pygame.sprite.Group(),        # group for all bomb sprites
        "hero"  : pygame.sprite.GroupSingle()   # single group for the hero
    }
    
    # Game score (0: actual score, 1: maximum score possible)
    SCORE = [0, 0]
    
    # Initialize window
    PARAM["screen"] = pygame.display.set_mode((PARAM["ww"], PARAM["wh"]))
    PARAM["background"] = pygame.Surface(PARAM["screen"].get_size()).convert()
    pygame.display.set_caption("MMP Blatt 6 - Aufgabe 1")
    pygame.mouse.set_visible(True)
    
    # Shortcut for Python's random function
    def rand(min, max):
        return random.randint(min, max)
    
    # Helper function that limits a given number to the screen dimensions
    def limit(num, dimension):
        return max(0, min(num, dimension))
    
    # Sprite for moving balls
    class Ball(pygame.sprite.Sprite):
        # Create a new ball sprite
        def __init__(self, click_position):
            pygame.sprite.Sprite.__init__(self)
            # Create ball
            self.radius = rand(PARAM["ball_min"], PARAM["ball_max"])
            self.image = pygame.Surface((self.radius*2, self.radius*2), pygame.SRCALPHA, 32)
            self.image = self.image.convert_alpha()
            pygame.draw.circle(self.image, pygame.Color(rand(0, 255), rand(0, 255), rand(0, 255)), (self.radius, self.radius), self.radius)
            # Correct given position to keep the entire ball within the borders of the window
            position = [0, 0]
            for i in [0, 1]:
                position[i] = limit(click_position[i]-self.radius, (PARAM["ww"] if i == 0 else PARAM["wh"])-PARAM["hero_size"])
            # Set ball's position
            self.rect = self.image.get_rect()
            self.rect.topleft = ((min(position[0], PARAM["ww"]-self.radius), min(position[1], PARAM["wh"]-self.radius)))
            self.pos = [float(self.rect.left), float(self.rect.top)]
            # Movement vector (direction), speed = length of vector
            self.movement = [rand(-PARAM["ww"], PARAM["ww"]), rand(-PARAM["wh"], PARAM["wh"])]
        # Update the ball's position
        def update(self, factor):
            for i in [0, 1]:
                self.pos[i] += factor * self.movement[i]
                # Check if ball has reached the borders, mirror the movement
                max = PARAM["ww"] if i == 0 else PARAM["wh"]
                if self.pos[i] <= 0:
                    self.pos[i] = 0
                    self.movement[i] = -self.movement[i]
                elif self.pos[i] >= max - self.radius*2:
                    self.pos[i] = max - self.radius*2
                    self.movement[i] = -self.movement[i]
            # Update the actual position of the sprite
            self.rect.left = round(self.pos[0])
            self.rect.top  = round(self.pos[1])
    
    # Sprite for bombs
    class Bomb(pygame.sprite.Sprite):
        # Create a new bomb
        def __init__(self, position):
            pygame.sprite.Sprite.__init__(self)
            # Create bomb
            self.image = pygame.Surface((PARAM["bomb_size"]*2, PARAM["bomb_size"]*2), pygame.SRCALPHA, 32)
            self.image = self.image.convert_alpha()
            pygame.draw.circle(self.image, PARAM["bomb_color"], (PARAM["bomb_size"], PARAM["bomb_size"]), PARAM["bomb_size"])
            # Set bomb's position
            self.rect = self.image.get_rect()
            self.rect.topleft = position
            self.pos = [position[0], position[1]]
        # Update the bomb's position
        def update(self, factor):
            self.pos[1] -= factor * PARAM["wh"] * 1.5
            if self.pos[1] <= 0:
                # Remove the bomb if it leaves the window
                SPRITES["bombs"].remove(self)
            else:
                self.rect.top = round(self.pos[1])
    
    # Sprite for the hero
    class Hero(pygame.sprite.Sprite):
        # Create the hero
        def __init__(self):
            pygame.sprite.Sprite.__init__(self)
            # Create simple rectangle for hero
            self.image = pygame.Surface((PARAM["hero_size"], PARAM["hero_size"]))
            self.image.fill(PARAM["hero_color"])
            self.pos = [PARAM["ww"]/2 - PARAM["hero_size"]/2, PARAM["wh"] - PARAM["hero_size"]]
            self.rect = self.image.get_rect()
            self.rect.topleft = (round(self.pos[0]), round(self.pos[1]))
            # Set hero not moving
            self.moving = False
        # Update the hero's position
        def update(self, factor):
            if (self.moving != False):
                # Calculate new position (limited to window dimensions) and set position
                movement = factor * PARAM["ww"] / 2
                if (self.moving == K_LEFT):
                    self.pos[0] -= movement
                elif (self.moving == K_RIGHT):
                    self.pos[0] += movement
                self.pos[0] = limit(PARAM["ww"]-PARAM["hero_size"], self.pos[0])
                self.rect.left = round(self.pos[0])
        # Start moving the hero in the given direction
        def move(self, direction):
            self.moving = direction
        # Stop the hero's movement
        def stop(self):
            self.moving = False
    
    # Function that checks for collisions between balls and bombs and removes the elements if a collision was detected
    def collision_detection():
        # Iterate through all bombs and check collision with each ball
        for bomb in SPRITES["bombs"]:
            for ball in SPRITES["balls"]:
                if pygame.sprite.collide_circle(bomb, ball) == True:
                    # We have a collision, so remove both the bomb and the ball from the sprite groups
                    SCORE[0] += PARAM["hit_points"]
                    SPRITES["bombs"].remove(bomb)
                    SPRITES["balls"].remove(ball)
                    break
    
    # Function that adds a new ball to the game
    def add_new_ball(position):
        SCORE[1] += PARAM["hit_points"] + PARAM["bomb_points"]
        SPRITES["balls"].add(Ball(position))
    
    # Function that shoots a new bomb
    def shoot_bomb():
        SCORE[0] += PARAM["bomb_points"]
        SPRITES["bombs"].add(Bomb((SPRITES["hero"].sprite.pos[0] + PARAM["hero_size"]/2, PARAM["wh"] - PARAM["hero_size"])))
    
    # Function that draws one frame
    def draw_frame(time):
        factor = time / PARAM["speed_factor"]
        # Draw background
        PARAM["background"].fill(PARAM["bg_color"])
        PARAM["screen"].blit(PARAM["background"], (0, 0))
        # Update balls, bombs and hero, do collision detection, then draw
        for i in SPRITES:
            SPRITES[i].update(factor)
        collision_detection()
        for i in SPRITES:
            SPRITES[i].draw(PARAM["screen"])
        # Finally draw score
        text = PARAM["score_font"].render(str(SCORE[0]) + " / " + str(SCORE[1]), True, PARAM["score_color"])
        PARAM["screen"].blit(text, (15, 10))
        pygame.display.update()
        
    # Add initial elements (hero and balls) to the game
    SPRITES["hero"].add(Hero())
    for i in [0, PARAM["init_balls"]-1]:
        add_new_ball((rand(0, PARAM["ww"]), rand(0, PARAM["wh"])))
    
    # Return the functions that may be used publicly
    return draw_frame, add_new_ball, SPRITES["hero"].sprite.move, SPRITES["hero"].sprite.stop, shoot_bomb


# Initialize the game and get public functions from the closure
draw, add_ball, move_hero, stop_hero, shoot = GAME()

# Create a clock for time-based animation
clock = pygame.time.Clock()

# Main loop
while True:
    # Draw a new frame
    draw(clock.tick())
    
    # Event handler
    for event in pygame.event.get():
        if event.type == MOUSEBUTTONDOWN:
            # Mouse click handler, adds a new ball
            add_ball(pygame.mouse.get_pos())
        elif event.type == KEYDOWN:
            # Key handler
            if event.key in [K_LEFT, K_RIGHT]:
                # Move hero left or right
                move_hero(event.key)
            elif event.key == K_SPACE:
                # Shoot a new bomb
                shoot()
        elif event.type == KEYUP:
            # Key handler to stop the hero's movement
            if event.key in [K_LEFT, K_RIGHT]:
                stop_hero()
        elif event.type == QUIT:
            sys.exit()